// Copyright (C) Maxime MORGE, 2024
package org.scata.algorithm
import org.scata.core._

/**
 * The Consensus-Based Auction Algorithm (CBAA) is
 * a single-assignment strategy
 */
class CBAA(pb : SingleAssignmentProblem) {
  private val debug = true

  // Generate a fully connected communication network
  private val neighbours: Map[Worker, List[Worker]] = pb.workers.toList.map { worker =>
    worker -> pb.workers.toList // Each worker is connected to all workers, including themselves
  }.toMap

  private var workerTaskList: Map[Worker, Map[Task, Boolean]] = pb.workers.toList.map { worker =>
    worker -> pb.tasks.toList.map { task =>
      task -> false
    }.toMap
  }.toMap

  private var winningBid: Map[Worker, Map[Task, (Worker, Int)]] = pb.workers.toList.map { worker =>
    worker -> pb.tasks.toList.map { task =>
      task -> (NoWorker, Int.MaxValue)
    }.toMap
  }.toMap


  /**
   * Returns true if the worker has no task assigned
   */
  private def isFree(worker : Worker) : Boolean = !pb.tasks.exists(task => workerTaskList(worker)(task))

  /**
   * CBAA Phase 1 for worker i
   * Selects the best task for the worker based on the cost
   * @param i index of the worker in the workers set
   * */
  private def selectTask(i : Int) : Unit = {
    val worker = pb.workers.toIndexedSeq(i)
    // Check if the worker has no tasks assigned
    if (isFree(worker)) {
      if (debug) println(s"Worker ${worker.name} is free")
      // Determine valid tasks based on cost comparison
      var validTasks: Map[Task, Boolean] = Map[Task, Boolean]()
      pb.tasks.foreach { task =>
        if (pb.cost(worker, task) < winningBid(worker)(task)._2) {
          if (debug) println(s"Worker ${worker.name} can perform task ${task.name}")
          validTasks = validTasks.updated(task, true)
        } else {
          validTasks = validTasks.updated(task, false)
        }
      }
      // If there are valid tasks, select the one with the minimum cost
      if (pb.tasks.exists(task => validTasks(task))) {
        val bestTask = validTasks.filter(_._2).keys.minBy(task => pb.cost(worker, task))
        if (debug) println(s"Worker ${worker.name} selects task ${bestTask.name} with cost ${pb.cost(worker, bestTask)}")
        workerTaskList = workerTaskList.updated(worker, workerTaskList(worker).updated(bestTask, true))
        winningBid = winningBid.updated(worker, winningBid(worker).updated(bestTask, (worker, pb.cost(worker, bestTask))))
      }
    }
  }

  /**
   * CBAA Phase 2 for worker i
   * Updates the winning bids of the worker based on the minimum winning bid
   * among its neighbors.
   * Returns true if the worker's winning bid is updated
   * @param i index of the worker in the workers set
   */
  private def consensus(i: Int): Boolean = {
    var isWinningBidUpdated = false
    val worker = pb.workers.toIndexedSeq(i)
    // Iterate over all tasks
    pb.tasks.foreach { task : Task =>
      // Find the minimum winning bid among the worker's neighbors (including itself)
      val minWinningBid = neighbours(worker)
        .map(neighbor => winningBid(neighbor)(task))
        .minBy(_._2)
      println(s"Worker ${worker.name} minimum winning bid for task ${task.name} is ${minWinningBid}")
      // Update the worker's winning bid for the task if a lower bid is found
      if (minWinningBid._2 < winningBid(worker)(task)._2) {
        println(s"Worker ${worker.name} updates winning bid for task ${task.name} to ${minWinningBid}")
        isWinningBidUpdated = true
        winningBid = winningBid.updated(worker, winningBid(worker).updated(task, minWinningBid))
        // Update the worker's task list to reflect the new winning bid
        workerTaskList = workerTaskList.updated(worker, workerTaskList(worker).updated(task, minWinningBid._1 == worker))
      }
    }
    isWinningBidUpdated
  }

  /**
   * Solves the single-assignment problem using the CBAA algorithm
   * @return the final assignment
   */
  def solve() : SingleAssignment = {
    var hasConverged = false
    while (!hasConverged) {
      if (debug) println("CBAA Phase 1")
      for (i <- 0 until pb.m) {
        selectTask(i)
      }
      hasConverged = true
      if (debug) println("CBAA Phase 2")
      for (i <- 0 until pb.m) {
        if (consensus(i)) {
          hasConverged = false
        }
        if (debug) println(s"CBAA Phase 2 has converged: ${hasConverged}")
      }
    }
    // Generate the final assignment
    val assignment = new SingleAssignment(pb)
    pb.workers.foreach { worker =>
      pb.tasks.foreach { task =>
        if (workerTaskList(worker)(task)) {
          assignment.bundle = assignment.bundle.updated(worker, task)
        }
      }
    }
    return assignment
  }
}
