// Copyright (C) Maxime MORGE, 2024
package org.scata.core

import scala.collection.SortedSet

/**
 * Class representing an allocation as
 * a single-assignment of the tasks to some workers.
 * @param pb is a single-assignment instance
 */
class SingleAssignment(val pb: SingleAssignmentProblem) {

  var bundle: Map[Worker, Task] = Map[Worker, Task]()

  pb.workers.foreach{
    worker=> bundle += worker -> NoTask
  }

  override def toString: String =
    pb.workers.toList.map(worker => s"$worker: ${bundle(worker)}").mkString("\n")

  override def equals(that: Any): Boolean =
    that match {
      case that: SingleAssignment => that.canEqual(this) && this.bundle == that.bundle
      case _ => false
    }

  override def hashCode(): Int = this.bundle.hashCode()

  def canEqual(a: Any): Boolean = a.isInstanceOf[SingleAssignment]

  /**
   * Returns a copy
   */
  @throws(classOf[RuntimeException])
  private def copy(): SingleAssignment = {
    val assignment = new SingleAssignment(pb)
    this.bundle.foreach {
      case (worker: Worker, task: Task) =>
        assignment.bundle = assignment.bundle.updated(worker, task)
      case _ => throw new RuntimeException("Not able to copy bundle")
    }
    assignment
  }

  /**
   * Updates an assignment with a new bundle for a computing node
   */
  def update(worker: Worker, task: Task): SingleAssignment = {
    val allocation = this.copy()
    allocation.bundle = allocation.bundle.updated(worker, task)
    allocation
  }

  /**
   * Returns true if each task is assigned to no more than one worker
   */
  private def isPartition: Boolean = {
    val tasks = bundle.values.toList
    tasks.distinct.size == tasks.size
  }

  /**
   * Returns true if the assignment is complete, i.e., every task is assigned to a worker
   */
  private def isComplete: Boolean = {
    pb.tasks.forall(task => bundle.values.toSet.contains(task))
  }

  /**
   * Returns true if the allocation is sound, i.e a complete partition of the tasks
   */
  def isSound: Boolean = isPartition && isComplete

  /**
   * Returns the worker which has the task in its bundle
   */
  def worker(task: Task): Option[Worker] = {
    bundle.find {
      case (_, t) if task.equals(t) => true
      case _ => false
    }.map(_._1)
  }
}