// Copyright (C) Maxime MORGE, 2024
package org.scata.core

import scala.collection.SortedSet
import org.scata.utils.RandomUtils

/**
  * Class representing a Single Assignment Problem
  *
  * @param workers are the workers
  * @param tasks are the tasks
  */
class SingleAssignmentProblem(val workers: SortedSet[Worker],
                              val tasks : SortedSet[Task],
                              val cost : Map[(Worker,Task), Int]) {

  /**
    * Returns a string describing the MASTAPlus problem
    */
  override def toString: String = {
    val workersStr = workers.map(_.name).mkString(", ")
    val tasksStr = tasks.map(_.name).mkString(", ")

    val costStr = workers.map { worker =>
      val taskCosts = tasks.map { task =>
        s"${task.name}: ${cost((worker, task))}"
      }.mkString(", ")
      s"${worker.name} -> [$taskCosts]"
    }.mkString("\n")

    s"Workers (${workers.size}): $workersStr\n" +
      s"Tasks (${tasks.size}): $tasksStr\n" +
      s"Costs:\n$costStr"
  }

  /**
    * Returns the number of computing nodes
    */
  def m: Int = workers.size

  /**
    * Returns the number of tasks
    */
  def n: Int = tasks.size
}

/**
  * Factory for SingleAssignmentProblem
  */
object SingleAssignmentProblem{

  implicit val order  : Ordering[Double] = Ordering.Double.TotalOrdering
  // eventually Ordering.Double.IeeeOrdering

  private val MAX_COST : Int = 100 // Maximum task cost

  /**
    * Returns a random single-assignment problem  instance with
    * @param m nodes
    * @param n tasks
    */
  def randomProblem(m: Int, n: Int): SingleAssignmentProblem = {
    // Workers
    val workers: SortedSet[Worker] = collection.immutable.SortedSet[Worker]() ++
      (for (i <- 1 to m) yield new Worker(name = "w%02d".format(i)))

    // Tasks
    val tasks : SortedSet[Task] = collection.immutable.SortedSet[Task]() ++
      (for (i <- 1 to n) yield new Task(name = "t%02d".format(i)))

    var cost : Map[(Worker,Task),Int] = Map [(Worker,Task),Int]()
    // Adjust the resource sizes
    workers.foreach{ worker =>
      tasks.foreach{ task =>
          cost =  cost.updated((worker, task), RandomUtils.random(1, MAX_COST))
      }
    }
    new SingleAssignmentProblem(workers, tasks, cost)
  }
}
